@@ -1,28 +1,39 @@
 # -*- coding: utf-8 -*-
 """Module that implements a basic command line interface (CLI) for AiiDAlab."""
-
 import json
+import logging
 import shutil
 from collections import defaultdict
-from contextlib import contextmanager
+from contextlib import contextmanager, nullcontext
+from dataclasses import asdict
 from fnmatch import fnmatch
 from pathlib import Path
-from textwrap import indent
-from textwrap import wrap
+from textwrap import indent, wrap
 
 import click
+import pkg_resources
+import requests_mock
 from click_spinner import spinner
-from packaging.requirements import InvalidRequirement
-from packaging.requirements import Requirement
+from jsonref import JsonRefError
+from jsonschema.exceptions import RefResolutionError, ValidationError
+from packaging.requirements import InvalidRequirement, Requirement
 from packaging.version import parse
 from tabulate import tabulate
 
 from . import __version__
 from .app import AppVersion
 from .app import _AiidaLabApp as AiidaLabApp
-from .config import AIIDALAB_APPS
-from .config import AIIDALAB_REGISTRY
-from .utils import load_app_registry_index
+from .config import AIIDALAB_APPS, AIIDALAB_REGISTRY
+from .fetch import fetch_from_url
+from .metadata import Metadata
+from .registry import build as build_registry
+from .utils import PEP508CompliantUrl, load_app_registry_index
+from .utils import parse_app_repo as _parse_app_repo
+
+logging.basicConfig(level=logging.INFO)
+
+
+SCHEMAS_CANONICAL_BASE_URL = "https://raw.githubusercontent.com/aiidalab/aiidalab/v21.10.0/aiidalab/registry/schemas"
 
 
 ICON_DETACHED = "\U000025AC"  # ▬
@@ -105,7 +116,10 @@ def search(app_query, prereleases):
         try:
             app_requirements = [Requirement(app_query)]
         except InvalidRequirement:  # interpreted as general search query
-            registry = load_app_registry_index()
+            try:
+                registry = load_app_registry_index()
+            except RuntimeError as error:
+                raise click.ClickException(error)
             app_requirements = [
                 Requirement(app_name)
                 for app_name in registry["apps"].keys()
@@ -162,22 +176,36 @@ def _parse_requirement(app_requirement):
     try:
         return Requirement(app_requirement)
     except InvalidRequirement as error:
+        from urllib.parse import urlsplit
+
+        if urlsplit(app_requirement).scheme and "@" not in app_requirement:
+            raise click.ClickException(
+                "It looks like you tried to install directly from a PEP 508 "
+                "compliant URL. Make sure to use the 'app-name @ url' syntax "
+                "to provide a name for the app."
+            )
         raise click.ClickException(
-            f"Invalid requirement '{app_requirement}': {error!s}"
+            f"Invalid requirement '{app_requirement}': {error!s}\n"
         )
 
 
-def _find_app_from_id(name):
+def _find_registered_app_from_id(name):
     """Find app for a given requirement."""
     try:
-        return AiidaLabApp.from_id(name)
+        app = AiidaLabApp.from_id(name)
+        if app.is_registered:
+            return app
+        else:
+            raise click.ClickException(
+                f"App '{app}' was installed locally and/or is not registered."
+            )
     except KeyError:
         raise click.ClickException(f"Did not find entry for app with name '{name}'.")
 
 
 def _find_app_and_releases(app_requirement):
     """Find app and a suitable release for a given requirement."""
-    app = _find_app_from_id(app_requirement.name)
+    app = _find_registered_app_from_id(app_requirement.name)
     matching_releases = app.find_matching_releases(app_requirement.specifier)
     return app, matching_releases
 
@@ -240,14 +268,38 @@ def show_environment(app_requirement, indent):
     click.echo(json.dumps(aggregated_environment, indent=indent))
 
 
-def _find_version_to_install(app_requirement, app, force, python_bin, prereleases):
+def _find_version_to_install(
+    app_requirement, force, dependencies, python_bin, prereleases
+):
+    if app_requirement.url is not None:
+        with fetch_from_url(app_requirement.url) as repo:
+            metadata = Metadata.parse(repo)
+            registry_entry = dict(name=app_requirement.name, metadata=asdict(metadata))
+            app = AiidaLabApp.from_id(
+                app_requirement.name, registry_entry=registry_entry
+            )
+            if not (force or (dependencies in ("install", "ignore"))):
+                raise click.ClickException(
+                    f"Unable to check compatibility for {app_requirement} prior to installation."
+                )
+
+        if force or not app.is_installed():
+            return app, PEP508CompliantUrl(app_requirement.url)
+        else:
+            return app, None
+
+    app = _find_registered_app_from_id(app_requirement.name)
+
+    assert dependencies in ("install", "require", "ignore")
     matching_releases = app.find_matching_releases(
         app_requirement.specifier, prereleases
     )
     compatible_releases = [
         version
         for version in matching_releases
-        if not python_bin or app.is_compatible(version, python_bin=python_bin)
+        if force
+        or (dependencies in ("install", "ignore"))
+        or app.is_compatible(version, python_bin=python_bin)
     ]
 
     # Best case scenario: matching and compatible release found!
@@ -289,13 +341,23 @@ def _find_version_to_install(app_requirement, app, force, python_bin, prerelease
     is_flag=True,
     help="Ignore all warnings and perform potentially dangerous operations anyways.",
 )
+@click.option(
+    "-d",
+    "--dependencies",
+    type=click.Choice(["install", "require", "ignore"]),
+    default="install",
+    help="Select whether to install app dependencies (the default), "
+    "require them to be installed prior to installation, "
+    "or completely ignore them.",
+    show_default=True,
+)
 @click.option(
     "--python",
     "python_bin",
-    default="python",
-    help="Specify the Python binary for which to check Python-dependencies for. "
-    "Set this option to an empty string to skip the dependency check.",
-    type=click.Path(dir_okay=False),
+    default=shutil.which("python"),
+    type=click.Path(dir_okay=False, exists=True),
+    help="The Python binary to use to install Python dependencies or to check their presence.",
+    show_default=True,
 )
 @click.option(
     "--pre",
@@ -303,7 +365,9 @@ def _find_version_to_install(app_requirement, app, force, python_bin, prerelease
     is_flag=True,
     help="Include prereleases among the candidates for installation.",
 )
-def install(app_requirement, yes, dry_run, force, python_bin, prereleases):
+def install(
+    app_requirement, yes, dry_run, force, dependencies, python_bin, prereleases
+):
     """Install apps.
 
     This command will install the latest version of an app matching the given requirement.
@@ -313,13 +377,18 @@ def install(app_requirement, yes, dry_run, force, python_bin, prereleases):
     Install the 'hello-world' app with the latest version that matches the specification '>=1.0':
 
         install hello-world>=1.0
-    """
 
+    Install the 'hello-world' app directly with a PEP 508 compliant URL:
+
+        install hello-world@git+https://github.com/aiidalab/aiidalab-hello-world.git
+
+    Note: It is necessary to explicitly specify the app name before the '@'.
+    """
     with _spinner_with_message("Collecting apps matching requirements... "):
         install_candidates = {
             requirement: _find_version_to_install(
                 requirement,
-                _find_app_from_id(requirement.name),
+                dependencies=dependencies,
                 force=force,
                 python_bin=python_bin,
                 prereleases=prereleases,
@@ -332,7 +401,7 @@ def install(app_requirement, yes, dry_run, force, python_bin, prereleases):
         return
 
     elif any(version is None for (_, version) in install_candidates.values()):
-        for requirement, (_, version) in sorted(install_candidates.items()):
+        for requirement, (_, version) in install_candidates.items():
             if version is None:
                 click.secho(
                     f"Requirement '{requirement}' already satisfied.",
@@ -348,7 +417,7 @@ def install(app_requirement, yes, dry_run, force, python_bin, prereleases):
     apps_table = tabulate(
         [
             [app.name, version, app.path]
-            for _, (app, version) in sorted(install_candidates.items())
+            for _, (app, version) in install_candidates.items()
             if version is not None
         ],
         headers=["App", "Version", "Path"],
@@ -362,10 +431,19 @@ def install(app_requirement, yes, dry_run, force, python_bin, prereleases):
                         f"Would install '{app.name}' version '{version}'.", fg="green"
                     )
                 else:
-                    app.install(version=version)
-                    click.secho(
-                        f"Installed '{app.name}' version '{version}'.", fg="green"
-                    )
+                    try:
+                        app.install(
+                            version=version,
+                            install_dependencies=dependencies == "install",
+                            python_bin=python_bin,
+                        )
+                    except RuntimeError as error:
+                        click.secho(f"Error: {error}", fg="red")
+                        break
+                    else:
+                        click.secho(
+                            f"Installed '{app.name}' version '{version}'.", fg="green"
+                        )
 
 
 @cli.command()
@@ -463,5 +541,163 @@ def uninstall(app_name, yes, dry_run, force):
                     )
 
 
+@contextmanager
+def _mock_schemas_endpoints():
+    schema_paths = [
+        path
+        for path in pkg_resources.resource_listdir(f"{__package__}.registry", "schemas")
+        if path.endswith(".schema.json")
+    ]
+
+    with requests_mock.Mocker(real_http=True) as mocker:
+        for schema_path in schema_paths:
+            schema = json.loads(
+                pkg_resources.resource_string(
+                    f"{__package__}.registry", f"schemas/{schema_path}"
+                )
+            )
+            mocker.get(
+                schema.get("$id", f"{SCHEMAS_CANONICAL_BASE_URL}/{schema_path}"),
+                text=json.dumps(schema),
+            )
+        yield
+
+
+@cli.group()
+def registry():
+    """Functions related to managing an app registry."""
+
+
+@registry.command()
+@click.argument("url")
+def parse_app_repo(url):
+    """Parse an app repo for metadata and other information.
+
+    Use this command to parse a local or remote app repository for the app
+    metadata and environment specification.
+
+    Examples:
+
+    For a local app repository, provide the absolute or relative path:
+
+        parse-app-repo /path/to/aiidalab-hello-world
+
+    For a remote app repository, provide a PEP 508 compliant URL, for example:
+
+        parse-app-repo git+https://github.com/aiidalab/aiidalab-hello-world.git@v1.0.0
+    """
+    click.echo(f"Parsing {url} ...", err=True)
+    try:
+        click.echo(json.dumps(_parse_app_repo(url)))
+    except (ValueError, TypeError) as error:
+        click.secho(
+            f"Failed to parse metadata from '{url}': {error!s}",
+            err=True,
+            fg="red",
+        )
+
+
+@registry.command()
+@click.option(
+    "--apps",
+    type=click.Path(exists=True, dir_okay=False),
+    default="apps.yaml",
+    help="Path to the registry's apps.yaml file.",
+    show_default=True,
+)
+@click.option(
+    "--categories",
+    type=click.Path(exists=True, dir_okay=False),
+    default="categories.yaml",
+    help="Path to the registry's categories.yaml file.",
+    show_default=True,
+)
+@click.option(
+    "-o",
+    "--out",
+    type=click.Path(file_okay=False, writable=True),
+    default="build/",
+    show_default=True,
+    help="Path to the base directory of where the registry's HTML and API pages are generated.",
+)
+@click.option(
+    "--html-path",
+    type=str,
+    help="Relative path to the build directory at which the HTML pages are generated. "
+    "Set to an empty value to skip the build of HTML pages entirely.",
+    default=".",
+    show_default=True,
+)
+@click.option(
+    "--api-path",
+    type=str,
+    help="Relative path to the build directory at which the API pages are generated. "
+    "Set to an empty value to skip the build of API pages entirely.",
+    default="api/v1",
+    show_default=True,
+)
+@click.option("--static", type=click.Path(file_okay=False, exists=True))
+@click.option("-t", "--templates", type=click.Path(file_okay=False, exists=True))
+@click.option(
+    "--validate/--no-validate",
+    is_flag=True,
+    default=True,
+    show_default=True,
+    help="Validate inputs and outputs against the published or local schemas.",
+)
+@click.option(
+    "-m",
+    "--mock-schemas-endpoints",
+    "mock_schemas",
+    is_flag=True,
+    help="Mock the schemas endpoints such that the local versions are used insted of the published ones.",
+)
+def build(
+    apps,
+    categories,
+    out,
+    html_path,
+    api_path,
+    static,
+    templates,
+    validate,
+    mock_schemas,
+):
+    """Build the app store website and API endpoints.
+
+    Example:
+
+        build --apps=apps.yaml --categories=categories.yaml --out=./build/
+    """
+    if any(path and Path(path).is_absolute() for path in (html_path, api_path)):
+        raise click.ClickException("The html- and api-paths must be relative paths.")
+
+    maybe_mock = _mock_schemas_endpoints() if mock_schemas else nullcontext()
+    with maybe_mock:
+        try:
+            build_registry(
+                apps_path=Path(apps),
+                categories_path=Path(categories),
+                base_path=Path(out),
+                html_path=Path(html_path) if html_path else None,
+                api_path=Path(api_path) if api_path else None,
+                static_path=Path(static) if static else None,
+                templates_path=Path(templates) if templates else None,
+                validate_output=validate,
+                validate_input=validate,
+            )
+        except ValidationError as error:
+            raise click.ClickException(f"Error during schema validation:\n{error}")
+        except RefResolutionError as error:
+            raise click.ClickException(
+                f"Failed to resolve schema JSON reference: {error}\n\n"
+                "Maybe try to run with `--mock-schemas-endpoints` ?"
+            )
+        except JsonRefError as error:
+            raise click.ClickException(
+                f"Failed to resolve JSON reference: {error.reference}\n{error.cause}"
+            )
+
+
 if __name__ == "__main__":
     cli()
