@@ -2,6 +2,7 @@
 """Module to manage AiiDAlab apps."""
 
 import errno
+import io
 import logging
 import os
 import shutil
@@ -9,17 +10,15 @@ import sys
 import tarfile
 import tempfile
 from contextlib import contextmanager
-from dataclasses import dataclass
-from dataclasses import field
-from enum import Enum, auto
+from dataclasses import asdict, dataclass, field
+from enum import Enum, Flag, auto
 from itertools import repeat
 from pathlib import Path
+from subprocess import CalledProcessError
 from threading import Thread
 from time import sleep
 from typing import List
-from urllib.parse import urldefrag
-from urllib.parse import urlsplit
-from urllib.parse import urlunsplit
+from urllib.parse import urldefrag, urlsplit, urlunsplit
 from uuid import uuid4
 
 import requests
@@ -31,16 +30,24 @@ from watchdog.events import FileSystemEventHandler
 from watchdog.observers import Observer
 from watchdog.observers.polling import PollingObserver
 
-
 from .config import AIIDALAB_APPS
+from .environment import Environment
 from .git_util import GitManagedAppRepo as Repo
 from .git_util import git_clone
-from .utils import find_installed_packages
-from .utils import load_app_registry_entry
-from .utils import split_git_url
-from .utils import this_or_only_subdir
-from .utils import throttled
-
+from .metadata import Metadata
+from .utils import (
+    FIND_INSTALLED_PACKAGES_CACHE,
+    PEP508CompliantUrl,
+    find_installed_packages,
+    load_app_registry_entry,
+    load_app_registry_index,
+    run_pip_install,
+    run_reentry_scan,
+    run_verdi_daemon_restart,
+    split_git_url,
+    this_or_only_subdir,
+    throttled,
+)
 
 logger = logging.getLogger(__name__)
 
@@ -54,12 +61,21 @@ class AppVersion(Enum):
     NOT_INSTALLED = auto()
 
 
+class AppRemoteUpdateStatus(Flag):
+    NOT_REGISTERED = auto()
+    UP_TO_DATE = auto()
+    UPDATE_AVAILABLE = auto()
+    CANNOT_REACH_REGISTRY = auto()
+    DETACHED = auto()
+
+
 @dataclass
 class _AiidaLabApp:
 
     metadata: dict
     name: str
     path: Path
+    logo: str = field(default_factory=str)
     categories: List[str] = field(default_factory=list)
     releases: dict = field(default_factory=dict)
 
@@ -70,22 +86,58 @@ class _AiidaLabApp:
             **{
                 key: value
                 for key, value in registry_entry.items()
-                if key in ("categories", "metadata", "name", "releases")
+                if key in ("categories", "logo", "metadata", "name", "releases")
             },
         )
 
+    @staticmethod
+    def _migrate_registry_entry_to_v1(entry):
+        entry["logo"] = entry["metadata"].pop("logo", None)
+        entry["categories"] = entry["metadata"].pop("categories", None)
+        return entry
+
+    @classmethod
+    def _registry_entry_from_path(cls, path):
+        try:
+            entry = {
+                "name": path.stem,
+                "metadata": asdict(Metadata.parse(path)),
+                "releases": None,
+            }
+            return cls._migrate_registry_entry_to_v1(entry)
+        except TypeError:
+            logger.debug(f"Unable to parse metadata from '{path}'")
+            return {
+                "name": path.stem,
+                "metadata": dict(title=path.stem, description=""),
+                "releases": None,
+                "logo": None,
+                "categories": None,
+            }
+
     @classmethod
     def from_id(cls, app_id, registry_entry=None, apps_path=None):
-        if registry_entry is None:
-            registry_entry = load_app_registry_entry(app_id) or dict(
-                name=app_id, metadata=dict()
-            )
         if apps_path is None:
             apps_path = AIIDALAB_APPS
 
-        return cls.from_registry_entry(
-            path=Path(apps_path).joinpath(app_id), registry_entry=registry_entry
-        )
+        app_path = Path(apps_path).joinpath(app_id)
+
+        if registry_entry is None:
+            local_registry_entry = cls._registry_entry_from_path(app_path)
+            remote_registry_entry = load_app_registry_entry(app_id)
+            registry_entry = remote_registry_entry or local_registry_entry
+
+        return cls.from_registry_entry(path=app_path, registry_entry=registry_entry)
+
+    @property
+    def is_registered(self):
+        try:
+            app_registry_index = load_app_registry_index()
+        except RuntimeError as error:
+            logger.warning(str(error))
+            return None
+        else:
+            return self.name in app_registry_index["apps"]
 
     @property
     def _repo(self):
@@ -95,31 +147,79 @@ class _AiidaLabApp:
             return None
 
     def installed_version(self):
-        if self._repo:
-            try:
-                head_commit = self._repo.head().decode()
-                versions_by_commit = {
-                    split_git_url(release["url"])[1]: version
-                    for version, release in self.releases.items()
-                    if urlsplit(release["url"]).scheme.startswith("git+")
-                }
-                return versions_by_commit.get(head_commit, AppVersion.UNKNOWN)
-            except Exception as error:
-                logger.debug(f"Encountered error while determining version: {error}")
+        if self._repo and self.is_registered:
+            if self.dirty():
                 return AppVersion.UNKNOWN
+            else:
+                try:
+                    head_commit = self._repo.head().decode()
+                    versions_by_commit = {
+                        split_git_url(release["url"])[1]: version
+                        for version, release in self.releases.items()
+                        if urlsplit(release["url"]).scheme.startswith("git+")
+                    }
+                    return versions_by_commit.get(
+                        head_commit, self.metadata.get("version", AppVersion.UNKNOWN)
+                    )
+                except Exception as error:
+                    logger.debug(
+                        f"Encountered error while determining version: {error}"
+                    )
+                    return AppVersion.UNKNOWN
         elif self.path.exists():
-            return AppVersion.UNKNOWN
+            return self.metadata.get("version", AppVersion.UNKNOWN)
         return AppVersion.NOT_INSTALLED
 
+    def available_versions(self):
+        if self.is_registered:
+            yield from sorted(self.releases, key=parse, reverse=True)
+
     def dirty(self):
         if self._repo:
             return self._repo.dirty()
 
-    def uninstall(self):
+    def is_installed(self):
+        """The app is installed if the corresponding folder is present."""
+        return self.path.exists()
+
+    def remote_update_status(self):
+        if self.is_installed():
+
+            # Check whether app is registered.
+            if self.is_registered is None:
+                return AppRemoteUpdateStatus.CANNOT_REACH_REGISTRY
+
+            if self.is_registered is False:
+                return AppRemoteUpdateStatus.NOT_REGISTERED
+
+            # Check whether the locally installed version is a registered release.
+            installed_version = self.installed_version()
+            if installed_version is AppVersion.UNKNOWN:
+                return AppRemoteUpdateStatus.DETACHED
+
+            # Check whether the locally installed version is the latest release.
+            available_versions = list(sorted(self.releases, key=parse, reverse=True))
+            if len(available_versions) and installed_version != available_versions[0]:
+                return AppRemoteUpdateStatus.UPDATE_AVAILABLE
+
+            # App must be up-to-date.
+            return AppRemoteUpdateStatus.UP_TO_DATE
+
+        return AppRemoteUpdateStatus(0)  # app is not installed
+
+    def _move_to_trash(self):
         trash_path = Path.home().joinpath(".trash", f"{self.name}-{uuid4()!s}")
         if self.path.exists():
             trash_path.parent.mkdir(parents=True, exist_ok=True)
             self.path.rename(trash_path)
+            return trash_path
+
+    def _restore_from(self, trash_path):
+        self._move_to_trash()
+        trash_path.rename(self.path)
+
+    def uninstall(self):
+        self._move_to_trash()
 
     def find_matching_releases(self, specifier, prereleases=None):
         matching_releases = list(
@@ -140,7 +240,13 @@ class _AiidaLabApp:
     def find_incompatibilities(self, version, python_bin=None):
         if python_bin is None:
             python_bin = sys.executable
-        environment = self.releases[version].get("environment")
+
+        if not self.is_registered:
+            assert version == self.installed_version()
+            environment = asdict(Environment.scan(self.path))
+        else:
+            environment = self.releases[version].get("environment")
+
         for key, spec in environment.items():
             if key == "python_requirements":
                 yield from zip(
@@ -153,6 +259,49 @@ class _AiidaLabApp:
     def is_compatible(self, version, python_bin=None):
         return not any(self.find_incompatibilities(version, python_bin))
 
+    def _install_dependencies(self, python_bin, stdout=None):
+        """Try to install the app dependencies with pip (if specified)."""
+
+        def _pip_install(*args, stdout=None):
+            # The implementation of this function is taken and adapted from:
+            # https://www.endpoint.com/blog/2015/01/getting-realtime-output-using-python/
+            stdout = stdout or sys.stdout
+
+            # Install package dependencies.
+            process = run_pip_install(*args, python_bin=python_bin)
+            for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
+                stdout.write(line)
+
+            # AiiDA plugins require reentry run to be found by AiiDA.
+            process = run_reentry_scan()
+            for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
+                stdout.write(line)
+
+            # Restarting the AiiDA daemon to import newly installed plugins.
+            process = run_verdi_daemon_restart()
+            for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
+                stdout.write(line)
+
+        for path in (self.path.joinpath(".aiidalab"), self.path):
+            if path.exists():
+                try:
+                    if path.joinpath("setup.py").is_file():
+                        _pip_install(
+                            "--use-feature=in-tree-build", str(path), stdout=stdout
+                        )
+                    elif path.joinpath("requirements.txt").is_file():
+                        _pip_install(
+                            f"--requirement={path.joinpath('requirements.txt')}",
+                            stdout=stdout,
+                        )
+                    else:
+                        logger.warning(
+                            f"Warning: App '{self.name}' does not declare any dependencies."
+                        )
+                    break
+                except CalledProcessError:
+                    raise RuntimeError("Failed to install dependencies.")
+
     def _install_from_path(self, path):
         if path.is_dir():
             shutil.copytree(this_or_only_subdir(path), self.path)
@@ -180,16 +329,24 @@ class _AiidaLabApp:
         base_url, ref = split_git_url(git_url)
         git_clone(base_url, ref, self.path)
 
-    def install(self, version=None):
+    def install(
+        self, version=None, python_bin=None, install_dependencies=True, stdout=None
+    ):
         if version is None:
             try:
                 version = list(sorted(self.releases, key=parse))[-1]
             except IndexError:
                 raise ValueError("No versions available for '{self}'.")
+        if python_bin is None:
+            python_bin = sys.executable
 
-        self.uninstall()
+        if isinstance(version, PEP508CompliantUrl):
+            url = version
+        else:
+            url = self.releases[version]["url"]
+
+        trash_path = self._move_to_trash()
 
-        url = self.releases[version]["url"]
         split_url = urlsplit(url)
         try:
             if split_url.scheme in ("", "file"):
@@ -208,11 +365,27 @@ class _AiidaLabApp:
                 raise NotImplementedError(
                     "Unsupported scheme: {split_url.scheme} ({url})"
                 )
+
+            # Install dependencies
+            if install_dependencies:
+                self._install_dependencies(python_bin, stdout=stdout)
         except RuntimeError as error:
-            raise RuntimeError(
-                f"Failed to install '{self.name}' (version={version}) at '{self.path}'"
-                f", due to error: {error}"
-            )
+            try:
+                if trash_path is None:
+                    # App, was not previously installed, just remove it.
+                    logger.info("Removing partially installed app.")
+                    self._move_to_trash()
+                else:
+                    # Attempt rollback to previous version.
+                    logger.info("Performing rollback to previously installed version.")
+                    self._restore_from(trash_path)
+            except RuntimeError as error:
+                logger.warning(f"Rollback failed due to error: {error}!")
+            finally:
+                raise RuntimeError(
+                    f"Failed to install '{self.name}' (version={version}) at '{self.path}'"
+                    f", due to error: {error}"
+                )
 
 
 class AppNotInstalledException(Exception):
@@ -362,7 +535,9 @@ class AiidaLabApp(traitlets.HasTraits):
     installed_version = traitlets.Union(
         [traitlets.Unicode(), traitlets.UseEnum(AppVersion)]
     )
-    updates_available = traitlets.Bool(readonly=True, allow_none=True)
+    remote_update_status = traitlets.UseEnum(
+        AppRemoteUpdateStatus, readonly=True, allow_none=True
+    )
     has_prereleases = traitlets.Bool()
     include_prereleases = traitlets.Bool()
 
@@ -378,6 +553,8 @@ class AiidaLabApp(traitlets.HasTraits):
         super().__init__()
 
         self.name = self._app.name
+        self.logo = self._app.logo
+        self.is_installed = self._app.is_installed
         self.path = str(self._app.path)
         self.refresh_async()
 
@@ -424,10 +601,6 @@ class AiidaLabApp(traitlets.HasTraits):
         # One should test what happens if the category won't be defined.
         return category in self._registry_data.categories
 
-    def is_installed(self):
-        """The app is installed if the corresponding folder is present."""
-        return os.path.isdir(self.path)
-
     def _has_git_repo(self):
         """Check if the app has a .git folder in it."""
         try:
@@ -436,19 +609,21 @@ class AiidaLabApp(traitlets.HasTraits):
         except NotGitRepository:
             return False
 
-    def install_app(self, version=None):
+    def install_app(self, version=None, stdout=None):
         """Installing the app."""
         with self._show_busy():
-            self._app.install(version=version)
+            self._app.install(version=version, stdout=stdout)
+            FIND_INSTALLED_PACKAGES_CACHE.clear()
             self.refresh()
             return self._installed_version()
 
-    def update_app(self, _=None):
+    def update_app(self, _=None, stdout=None):
         """Perform app update."""
         with self._show_busy():
             # Installing with version=None automatically selects latest
             # available version.
-            version = self.install_app(version=None)
+            version = self.install_app(version=None, stdout=stdout)
+            FIND_INSTALLED_PACKAGES_CACHE.clear()
             self.refresh()
             return version
 
@@ -459,12 +634,6 @@ class AiidaLabApp(traitlets.HasTraits):
             self._app.uninstall()
             self.refresh()
 
-    def _available_versions(self):
-        """Return all available and compatible versions."""
-        for version in sorted(self._app.releases, key=parse, reverse=True):
-            if self._is_compatible(version):
-                yield version
-
     def _installed_version(self):
         """Determine the currently installed version."""
         return self._app.installed_version()
@@ -479,19 +648,18 @@ class AiidaLabApp(traitlets.HasTraits):
             incompatibilities = dict(
                 self._app.find_incompatibilities(version=app_version)
             )
-            self.compatibility_info.update(
-                {
-                    app_version: [
-                        f"({eco_system}) {requirement}"
-                        for eco_system, requirement in incompatibilities.items()
-                    ]
-                }
-            )
+            self.compatibility_info = {
+                app_version: [
+                    f"({eco_system}) {requirement}"
+                    for eco_system, requirement in incompatibilities.items()
+                ]
+            }
+
             return not any(incompatibilities)
         except KeyError:
             return None  # compatibility indetermined for given version
 
-    def _updates_available(self):
+    def _remote_update_status(self):
         """Determine whether there are updates available.
 
         For this the app must be installed in a known version and there must be
@@ -505,20 +673,22 @@ class AiidaLabApp(traitlets.HasTraits):
         return False
 
     def _refresh_versions(self):
-        all_available_versions = list(self._available_versions())
-        self.has_prereleases = any(
-            parse(version).is_prerelease for version in all_available_versions
-        )
-        self.installed_version = self._installed_version()
+        self.installed_version = self._app.installed_version()
         self.include_prereleases = self.include_prereleases or (
             isinstance(self.installed_version, str)
             and parse(self.installed_version).is_prerelease
         )
-        self.available_versions = [
-            str(version)
-            for version in all_available_versions
-            if self.include_prereleases or not parse(version).is_prerelease
-        ]
+
+        all_available_versions = list(self._app.available_versions())
+        self.has_prereleases = any(
+            parse(version).is_prerelease for version in all_available_versions
+        )
+        if self._app.is_registered:
+            self.available_versions = [
+                version
+                for version in all_available_versions
+                if self.include_prereleases or not parse(version).is_prerelease
+            ]
 
     @throttled(calls_per_second=1)
     def refresh(self):
@@ -529,17 +699,13 @@ class AiidaLabApp(traitlets.HasTraits):
                 self.set_trait(
                     "compatible", self._is_compatible(self.installed_version)
                 )
-                if self.is_installed() and self._has_git_repo():
-                    self.installed_version = self._installed_version()
-                    modified = self._repo.dirty()
-                    self.set_trait(
-                        "detached",
-                        self.installed_version is AppVersion.UNKNOWN or modified,
-                    )
-                    self.set_trait("updates_available", self._updates_available())
-                else:
-                    self.set_trait("updates_available", None)
-                    self.set_trait("detached", None)
+                self.set_trait("remote_update_status", self._app.remote_update_status())
+                self.set_trait(
+                    "detached",
+                    (self.installed_version is AppVersion.UNKNOWN)
+                    if (self._has_git_repo() and self._app.is_registered)
+                    else None,
+                )
 
     def refresh_async(self):
         """Asynchronized (non-blocking) refresh of the app state."""
